from typing import Tuple, List
import pandas as pd
import numpy as np
from .raw import RawDataApi
from .basic import BasicDataApi
from .derived import DerivedDataApi
from .test_api import TestDataApi
from .research_api import ResearchApi
from ...util.singleton import Singleton
from ...util.calculator_item import *


class ResearchToolApi(metaclass=Singleton):
    
    def get_asset_info_by_key_words(self, key_words: Tuple[str] = ()):
        def get_info_list(dic):
            res = []
            for k,v in dic.items():
                res.append({
                    'asset_id':k,
                    'asset_name':v,
                })
            return res

        try:
            max_num = 20
            stock_info = RawDataApi().get_em_stock_info_by_key_words(key_words)
            stock_info = stock_info.head(min(max_num, stock_info.shape[0]))
            stock_dic = stock_info.set_index('stock_id')['name'].to_dict()

            index_info = BasicDataApi().get_index_info_by_key_words(key_words)
            index_info = index_info.head(min(max_num, index_info.shape[0]))
            index_dic = index_info.set_index('index_id')['desc_name'].to_dict()

            fund_info = BasicDataApi().get_fund_info_by_key_words(key_words=key_words)
            fund_info = fund_info.head(min(max_num, fund_info.shape[0]))
            fund_dic = fund_info.set_index('fund_id')['desc_name'].to_dict()

            fof_info = BasicDataApi().get_fof_info_by_key_words(manager_id='1',key_words=key_words)
            fof_info = fof_info.head(min(max_num, fof_info.shape[0]))
            fof_dic = fof_info.set_index('fof_id')['fof_name'].to_dict()

            stg_info = BasicDataApi().get_asset_info_by_type_by_key_words(asset_type=['策略指数'],key_words=key_words)
            stg_info = stg_info.head(min(max_num, stg_info.shape[0]))
            stg_dic = stg_info.set_index('real_id')['real_name'].to_dict()

            result = []
            res_i = get_info_list(stock_dic)
            if len(res_i) > 0:
                result.append({'asset_type':'股票','asset_list':res_i})
            res_i = get_info_list(index_dic)
            if len(res_i) > 0:
                result.append({'asset_type':'指数','asset_list':res_i})
            res_i = get_info_list(fund_dic)
            if len(res_i) > 0:
                result.append({'asset_type':'公募基金','asset_list':res_i})
            res_i = get_info_list(fof_dic)
            if len(res_i) > 0:
                result.append({'asset_type':'私募基金','asset_list':res_i})
            res_i = get_info_list(stg_dic)
            if len(res_i) > 0:
                result.append({'asset_type':'策略指数','asset_list':res_i})
            return result

        except Exception as e:
            print(f'failed to asset price <err_msg> {e} from ResearchToolApi.get_asset_info_by_key_words')
            return []

    def get_stats_df(self, df, begin_date_dic):
        stats_name = {
                'start_date':'成立日',
                'annual_ret':'年化收益',
                'annual_vol':'年化波动',
                'sharpe':'夏普',
                'mdd':'最大回撤',
                'mdd_date1':'最大回撤开始日',
                'mdd_date2':'最大回撤结束日',
            }
        res = []
        for asset in df:
            _df = df[[asset]].copy().dropna()
            stats_dic = CalculatorBase.get_stat_result(dates=_df.index, values=_df[asset], risk_free_rate=0.025)
            stats_dic['资产名'] = asset
            _asset = asset.split('/')[0]
            if _asset in begin_date_dic:
                stats_dic['start_date'] = begin_date_dic[_asset]
            res.append(stats_dic)
        stats_df = pd.DataFrame(res).set_index('资产名')[stats_name.keys()].rename(columns=stats_name)
        stats_df['成立日'] = stats_df['成立日'].astype(str)
        stats_df['最大回撤开始日'] = stats_df['最大回撤开始日'].astype(str)
        stats_df['最大回撤结束日'] = stats_df['最大回撤结束日'].astype(str)
        stats_df.columns.name = '价格比指标'
        return stats_df

    def get_all_asset_price(self, stock_info, index_info, fund_info, hf_fund_info, hf_index_info, asset_id):
        raw = RawDataApi()
        basic = BasicDataApi()
        id_list_1 = stock_info[stock_info.stock_id.isin(asset_id)].stock_id.tolist()
        id_list_2 = index_info[index_info.index_id.isin(asset_id)].index_id.tolist()
        id_list_3 = fund_info[fund_info.fund_id.isin(asset_id)].fund_id.tolist()
        id_list_4 = hf_fund_info[hf_fund_info.fof_id.isin(asset_id)].fof_id.tolist()
        id_list_5 = hf_index_info[hf_index_info.real_id.isin(asset_id)].real_id.tolist()
        result = []
        if len(id_list_1) > 0 :
            # stock price
            dic = stock_info.set_index('stock_id')['name'].to_dict()
            df = raw.get_em_stock_post_price(stock_list=id_list_1)
            df = df.pivot_table(index='datetime',columns='stock_id',values='close')
            #df = df.rename(columns=dic)
            if not df.empty:
                result.append(df)
        if len(id_list_2) > 0:        
            # index price
            dic = index_info.set_index('index_id')['desc_name'].to_dict()
            df = basic.get_index_price(index_list=id_list_2)
            df = df.pivot_table(index='datetime',columns='index_id',values='close')
            #df = df.rename(columns=dic)
            if not df.empty:
                result.append(df)
        if len(id_list_3) > 0:
            # fund nav
            dic = fund_info.set_index('fund_id')['desc_name'].to_dict()
            df = basic.get_fund_nav(fund_list=id_list_3)
            df = df.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value')
            #df = df.rename(columns=dic)
            if not df.empty:
                result.append(df)
        if len(id_list_4) > 0:
            # hf fund nav
            dic = hf_fund_info.set_index('fof_id')['fof_name'].to_dict()
            default_manager_id = 'py1'
            fund_nav1 = DerivedDataApi().get_fof_nav(fof_id_list=id_list_4,manager_id=default_manager_id)
            col1 = []
            col2 = []
            _result = []
            if not fund_nav1.empty:
                fund_nav1 = fund_nav1.pivot_table(index='datetime',columns='fof_id',values='adjusted_nav')
                col1 = fund_nav1.columns.tolist()
                _result.append(fund_nav1[col1])
            fund_nav2 = DerivedDataApi().get_fof_nav_public(fof_id_list=id_list_4)
            if fund_nav2 is not None and not fund_nav2.empty:
                fund_nav2 = fund_nav2.pivot_table(index='datetime',columns='fof_id',values='adjusted_nav')
                col2 = fund_nav2.columns.tolist()
                col2 = [_ for _ in col2 if _ not in col1]
                _result.append(fund_nav2[col2])
            if len(_result) > 0:
                df = pd.concat(_result, axis=1).sort_index()
                #df = df.rename(columns=dic)
                df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample('W-FRI').last()
                df.index = [i.date() for i in df.index]
                if not df.empty:
                    result.append(df)
        if len(id_list_5) > 0:
            # hf_index_price
            dic = hf_index_info.set_index('real_id')['real_name'].to_dict()
            df = raw.get_hf_index_price(index_ids=id_list_5)
            df = df.pivot_table(index='index_date', columns='index_id', values='close')
            #df = df.rename(columns=dic)
            if not df.empty:
                result.append(df)
        return pd.concat(result,axis=1).sort_index()

    def get_all_asset_price_v2(self, stock_info, index_info, fund_info, hf_fund_info, hf_index_info, diy_index_info,  asset_id):
        raw = RawDataApi()
        basic = BasicDataApi()
        id_list_1 = stock_info[stock_info.stock_id.isin(asset_id)].stock_id.tolist()
        id_list_2 = index_info[index_info.index_id.isin(asset_id)].index_id.tolist()
        id_list_3 = fund_info[fund_info.fund_id.isin(asset_id)].fund_id.tolist()
        id_list_4 = hf_fund_info[hf_fund_info.RECORD_CD.isin(asset_id)].RECORD_CD.tolist()
        id_list_5 = hf_index_info[hf_index_info.real_id.isin(asset_id)].real_id.tolist()
        id_list_6 = diy_index_info[diy_index_info.index_id.isin(asset_id)].index_id.tolist()

        result = []
        if len(id_list_1) > 0 :
            # stock price
            dic = stock_info.set_index('stock_id')['name'].to_dict()
            df = raw.get_em_stock_post_price(stock_list=id_list_1)
            df = df.pivot_table(index='datetime',columns='stock_id',values='close')
            #df = df.rename(columns=dic)
            if not df.empty:
                result.append(df)
        if len(id_list_2) > 0:        
            # index price
            dic = index_info.set_index('index_id')['desc_name'].to_dict()
            df = basic.get_index_price(index_list=id_list_2)
            df = df.pivot_table(index='datetime',columns='index_id',values='close')
            #df = df.rename(columns=dic)
            if not df.empty:
                result.append(df)
        if len(id_list_3) > 0:
            # fund nav
            dic = fund_info.set_index('fund_id')['desc_name'].to_dict()
            df = basic.get_fund_nav(fund_list=id_list_3)
            df = df.pivot_table(index='datetime',columns='fund_id',values='adjusted_net_value')
            #df = df.rename(columns=dic)
            if not df.empty:
                result.append(df)
        if len(id_list_4) > 0:
            # hf fund nav
            dic = hf_fund_info.set_index('RECORD_CD')['SEC_FULL_NAME'].to_dict()
            df1 = ResearchApi().get_pf_nav(record_cds=id_list_4, source='tl')
            df1 = df1.pivot_table(index='END_DATE',columns='RECORD_CD',values='ADJ_NAV')
            df2 = ResearchApi().get_pf_nav(record_cds=id_list_4, source='py')
            if not df2.empty:
                df2 = df2.pivot_table(index='END_DATE',columns='RECORD_CD',values='ADJ_NAV')
                common_list = df1.columns.intersection(df2.columns).tolist()
                _df1 = df1[df1.columns[~df1.columns.isin(common_list)]]
                if _df1.empty:
                    df = df1
                else:
                    df = pd.concat([df1,df2],axis=1)
            else:
                df = df1
            if not df.empty:
                result.append(df)
        if len(id_list_5) > 0:
            # hf_index_price
            dic = hf_index_info.set_index('real_id')['real_name'].to_dict()
            df = raw.get_hf_index_price(index_ids=id_list_5)
            df = df.pivot_table(index='index_date', columns='index_id', values='close')
            #df = df.rename(columns=dic)
            if not df.empty:
                result.append(df)
        if len(id_list_6) > 0:
            # diy stg index
            dic = diy_index_info.set_index('index_id')['desc_name'].to_dict()
            df = TestDataApi().get_stg_index_price(index_list=id_list_6)
            df = df.pivot_table(index='datetime', columns='index_id', values='close')
            #df = df.rename(columns=dic)
            if not df.empty:
                result.append(df)
        df = pd.concat(result,axis=1).sort_index()
        df.index.name = 'datetime'
        return df

    def get_data_dict_of_list(self, df):
        dic = {}
        for col in df:
            dic[col] = df[col].values.tolist()
        return dic

    def get_asset_info_with_begin_date(self, key_words):
        max_num = 40
        stock_info = RawDataApi().get_em_stock_info_by_key_words(key_words)
        stock_info = stock_info.head(min(max_num, stock_info.shape[0]))
        stock_dic = stock_info.set_index('stock_id')['name'].to_dict()

        index_info = BasicDataApi().get_index_info_by_key_words(key_words)
        index_info = index_info.head(min(max_num, index_info.shape[0]))
        index_dic = index_info.set_index('index_id')['desc_name'].to_dict()

        fund_info = BasicDataApi().get_fund_info_by_key_words(key_words=key_words)
        fund_info = fund_info.head(min(max_num, fund_info.shape[0]))
        fund_dic = fund_info.set_index('fund_id')['desc_name'].to_dict()

        hf_fund_info = ResearchApi().get_pf_info(name_list=key_words)
        hf_fund_info = hf_fund_info.head(min(max_num, hf_fund_info.shape[0]))
        hf_fund_dic = hf_fund_info.set_index('RECORD_CD')['SEC_FULL_NAME'].to_dict()

        stg_info = BasicDataApi().get_asset_info_by_type_by_key_words(asset_type=['策略指数'],key_words=key_words)
        stg_info = stg_info.head(min(max_num, stg_info.shape[0]))
        stg_dic = stg_info.set_index('real_id')['real_name'].to_dict()

        diy_stg_info = TestDataApi().get_stg_index_by_key_words(key_words)
        diy_stg_info = diy_stg_info.head(min(max_num, diy_stg_info.shape[0]))
        diy_stg_dic = diy_stg_info.set_index('index_id')['desc_name'].to_dict()


        stock_dic.update(index_dic)
        stock_dic.update(fund_dic)
        stock_dic.update(hf_fund_dic)
        stock_dic.update(stg_dic)
        stock_dic.update(diy_stg_dic)
        stock_dic = {k:v for k, v in stock_dic.items() if k != ''}
        price_df = self.get_all_asset_price_v2(stock_info, index_info, fund_info, hf_fund_info, stg_info, diy_stg_info, stock_dic.keys())
        _stock_dic = stock_info.set_index('stock_id')['name'].to_dict()
        res = []
        for col in price_df:
            if col in _stock_dic:
                asset_type = '股票'
            elif col in index_dic:
                asset_type = '指数'
            elif col in fund_dic:
                asset_type = '公募基金'
            elif col in hf_fund_dic:
                asset_type = '私募基金'
            elif col in stg_dic:
                asset_type = '私募指数'
            elif col in diy_stg_dic:
                asset_type = '自制指数'
            dic = {
                'id':col,
                'name':stock_dic.get(col,col),
                'begin_date':price_df[col].dropna().index[0],
                'asset_type':asset_type,
            }
            res.append(dic)
        df = pd.DataFrame(res)
        return df
    
    def super_price_ratio(self, asset_ids: Tuple[str]=(), bmk_id: str='', start_date=None, end_date=None):
        try:
            all_asset_ids = asset_ids if bmk_id in asset_ids else asset_ids + [bmk_id]
            stock_info = RawDataApi().get_em_stock_info(stock_list=all_asset_ids)
            index_info = BasicDataApi().get_index_info(index_list=all_asset_ids)
            fund_info = BasicDataApi().get_fund_info(fund_list=all_asset_ids)
            #hf_fund_info = BasicDataApi().get_fof_info(manager_id='1', fof_id_list=all_asset_ids)
            hf_fund_info = ResearchApi().get_pf_info(record_cd=all_asset_ids)
            hf_index_info = BasicDataApi().get_asset_info_by_real_id_and_asset_type(real_id_list=all_asset_ids,asset_type=['策略指数'])
            diy_stg_info = TestDataApi().get_stg_index_by_index_id(index_list=all_asset_ids)
            # id 和名字对照
            dic = stock_info.set_index('stock_id')['name'].to_dict()
            dic1 = index_info.set_index('index_id')['desc_name'].to_dict()
            dic2 = fund_info.set_index('fund_id')['desc_name'].to_dict()
            dic3 = hf_fund_info.set_index('RECORD_CD')['SEC_FULL_NAME'].to_dict()
            dic4 = hf_index_info.set_index('real_id')['real_name'].to_dict()
            dic5 = diy_stg_info.set_index('index_id')['desc_name'].to_dict()
            dic.update(dic1)
            dic.update(dic2)
            dic.update(dic3)
            dic.update(dic4)
            dic.update(dic5)
            name_dic = [{'asset_id':k,'asset_name':v} for k,v in dic.items()]

            price_df_ori = self.get_all_asset_price_v2(stock_info, index_info, fund_info, hf_fund_info, hf_index_info, diy_stg_info, asset_ids)
            price_df = price_df_ori.loc[start_date:end_date]
            benchmark_df_ori = self.get_all_asset_price_v2(stock_info, index_info, fund_info, hf_fund_info, hf_index_info, diy_stg_info, [bmk_id])
            benchmark_df = benchmark_df_ori.loc[start_date:end_date]
            price_df = pd.merge(price_df,benchmark_df, left_index=True, right_index=True, how='outer').sort_index()
            #price_df = price_df.join(benchmark_df).sort_index()
            price_df = price_df/price_df.bfill().iloc[0]
            #起始日统计
            begin_date_dic = {}
            for col in price_df_ori:
                begin_date_dic[col] = price_df_ori[col].dropna().index[0]
            for col in benchmark_df_ori:
                begin_date_dic[col] = benchmark_df_ori[col].dropna().index[0]
            benchmark_name = benchmark_df.columns[0]
            
            # 把资产价格起点设置在同期指数点位，形成散开方式
            _price_df = price_df.copy()
            asset_names = _price_df.columns.tolist()
            for asset_i in asset_names:
                b_d = _price_df[asset_i].dropna().index[0]
                ratio = price_df[benchmark_name].bfill().loc[b_d]
                _price_df[asset_i] = _price_df[asset_i] * ratio
            _price_df = _price_df / _price_df.bfill().iloc[0]
            # 各资产求价格比
            result = []
            for col in price_df:
                if col == benchmark_name:
                    continue
                _df = price_df[[col]].join(benchmark_df.rename(columns={benchmark_name:'benchmark'})).dropna()
                if _df.empty:
                    continue
                _df = _df / _df.iloc[0]
                _df = (_df[col] / _df['benchmark']).to_frame()
                _df.columns = [col]
                result.append(_df)
            df = pd.concat(result,axis=1).sort_index()
            cols = df.columns.tolist()
            cols = [f'{_}/{benchmark_name}' for _ in cols]
            df.columns= cols

            calculate_result = self.get_stats_df(df, begin_date_dic)
            whole_index = _price_df.index.union(df.index)
            whole_index = whole_index[(whole_index >= df.index[0]) & (whole_index <= df.index[-1])]
            _price_df = _price_df.reindex(whole_index).sort_index()
            df = df.reindex(whole_index).sort_index()
            result = {'mv':self.get_data_dict_of_list(_price_df),
                    'mv_diff':self.get_data_dict_of_list(df),
                    'table':calculate_result.T.to_dict(),
                    'name_dic':name_dic,
                    'datetime':df.index.astype(str).tolist(),
                    'mv_df':_price_df,
                    'mv_diff_df':df,
            }
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from ResearchToolApi.super_price_ratio')
            result = {'mv':{},
                    'mv_diff':{},
                    'table':{},
                    'name_dic':[],
                    'datetime':[],
            }
        return result

    def asset_allocation_backtest(self, index_list, index_wgt, frequency, start_date, end_date):
        try:
            frequency_dic = {
                'monthly' : DateOffset(months=1),
                'quarter' : DateOffset(months=3),
                'yearly' : DateOffset(months=12),
            }
            index_info = ResearchApi().get_index_info_by_id(index_list=index_list)
            index_name_dic = index_info.set_index('index_id')['desc_name'].to_dict()
            index_price = ResearchApi().get_index_quote(start_date=start_date,end_date=end_date,index_ids=index_list)
            index_price = index_price.pivot_table(index='datetime',columns='index_id',values='close').ffill().dropna()

            # get rebalance date list
            n = 0
            rebalance_dt = frequency_dic[frequency]
            begin_date = index_price.index[0]
            end_date = index_price.index[-1]
            rebalance_dt_list = []
            while(True):
                dt = (begin_date + rebalance_dt * n).date()
                if dt <= end_date:
                    rebalance_dt_list.append(dt)
                    n += 1
                else:
                    break
            # simple backtest
            result = []
            weight_res = []
            last_price = 1
            index_wgt_dic = {k: v for k, v in zip(index_list, index_wgt)}
            index_wgt_df = pd.DataFrame([index_wgt_dic])
            index_wgt_df = index_wgt_df[index_price.columns.tolist()]
            
            for idx, dt in enumerate(rebalance_dt_list):
                if dt == rebalance_dt_list[-1]:
                    next_dt = index_price.index[-1]
                else:
                    next_dt = rebalance_dt_list[idx + 1]
                    next_dt = index_price.loc[next_dt:].index[0]
                _index_price = index_price.loc[dt:next_dt]
                _index_price = _index_price / _index_price.iloc[0]
                index_mv = _index_price.mul(index_wgt_df.values)
                weight_i = index_mv.div(index_mv.sum(axis=1),axis=0)
                weight_res.append(weight_i)
                index_mv = index_mv.sum(axis=1).to_frame()
                index_mv = index_mv / index_mv.iloc[0] * last_price
                last_price = index_mv[0].iloc[-1]
                index_mv.columns = ['组合收益']
                result.append(index_mv)
            
            weight_df = pd.concat(weight_res)
            weight_df = weight_df.rename(columns=index_name_dic)
            port_mv = pd.concat(result).sort_index().drop_duplicates()
            index_price = index_price.rename(columns=index_name_dic)
            all_price = index_price.join(port_mv)
            all_price = all_price / all_price.iloc[0]

            # calculate portfolio status
            stats_name = {
                'start_date':'成立日',
                'annual_ret':'年化收益',
                'annual_vol':'年化波动',
                'sharpe':'夏普',
                'mdd':'最大回撤',
                'mdd_date1':'最大回撤开始日',
                'mdd_date2':'最大回撤结束日',
            }
            res = []
            index_wgt_dic_chn = { index_name_dic[k]: v  for k,v in index_wgt_dic.items()}
            for asset in all_price:

                stats_dic = CalculatorBase.get_stat_result(dates=all_price.index, values=all_price[asset], risk_free_rate=0.025)
                stats_dic['资产名'] = asset
                stats_dic['初始权重'] = index_wgt_dic_chn.get(asset, np.sum(index_wgt))
                res.append(stats_dic)
            port_stats = pd.DataFrame(res).set_index('资产名')[list(stats_name.keys())+['初始权重']].rename(columns=stats_name)
            round_3_list = ['年化收益','年化波动','夏普','最大回撤']
            for _ in round_3_list:
                port_stats[_] = port_stats[_].round(3)
            weight_df = weight_df[~weight_df.index.duplicated(keep='last')]
            result = {
                '组合收益': all_price,
                '各资产权重': weight_df,
                '回测指标': port_stats,
            }
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from ResearchToolApi.asset_allocation_backtest')
            result = {'组合收益':pd.DataFrame(),
                     '各资产权重':pd.DataFrame(),
                     '回测指标':pd.DataFrame(),
            }
        return result

if __name__ == '__main__':
    
    # 搜关键字返回资产名
    keys_words = ['茅台','贵州']
    result = ResearchToolApi().get_asset_info_by_key_words(key_words=keys_words)

    # 价格比计算
    asset_ids = ['600519.SH']
    bmk_id = 'hs300'
    start_date = datetime.date(2021,1,1)
    end_date = datetime.date(2022,1,1)
    result = ResearchToolApi().super_price_ratio(asset_ids=asset_ids,
                                                bmk_id=bmk_id,
                                                start_date=start_date,
                                                end_date=end_date,
                                                )



from typing import Tuple, Optional
import numpy as np
import pandas as pd
from surfing.data.api.raw import *
from surfing.data.api.basic import *
from surfing.data.api.derived import *

WENJIAN_FUND_LIST = ["SGM473","SNG191","SR9762","SNH765","SEW210","S3649A","SGM992","SJE335","SLC213","EE891B","SJJ077"]

class FOFApi(metaclass=Singleton):
    
    def get_pf_virtual_nav(self, fof_id='SLW695'):
        '''计算虚拟净值'''
        from surfing.data.manager.manager_fof_lite import FOFDataManagerLite
        try:
            fof_aa = BasicDataApi().get_fof_asset_allocation(manager_id='py1')
            fof_aa = fof_aa[fof_aa.asset_type == HoldingAssetType.HEDGE]
            df = FOFDataManagerLite._calc_virtual_net_value('py1', fof_id, list(fof_aa.fund_id.unique()))
            return df
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from {FOFApi.get_pf_virtual_nav}')

    def get_fof_pf_trade_history(self, fof_id_list=['SLW695'], manager_id='py1', fund_list=WENJIAN_FUND_LIST):
        '''获取FOF私募产品申赎记录'''
        from surfing.constant import HoldingAssetType, FOFTradeStatus
        try:
            with BasicDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                    FOFInfo
                ).filter(
                    FOFInfo.fof_id.in_(fund_list),
                    FOFInfo.manager_id == manager_id,
                )
                fund_info = pd.read_sql(query.statement, query.session.bind)
                rename_dict = fund_info[["fof_id","desc_name"]].set_index("fof_id").to_dict()["desc_name"]
                rename_dict["others_mv"] = "现金"

            fof_aa = BasicDataApi().get_fof_asset_allocation(manager_id=manager_id,fof_id_list=fof_id_list)
            fof_aa = fof_aa[(fof_aa.asset_type == HoldingAssetType.HEDGE) & (fof_aa.event_type.isin([FOFTradeStatus.PURCHASE, FOFTradeStatus.SUBSCRIBE, FOFTradeStatus.REDEEM]))]
            fof_aa['bs_flag'] = fof_aa.event_type.map(lambda x: -1 if x == FOFTradeStatus.REDEEM else 1)
            fof_aa['amount'] *= fof_aa.bs_flag
            fof_aa = fof_aa.pivot(index='datetime', columns='fund_id', values='amount').rename(columns = rename_dict)
            return fof_aa

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from {FOFApi.get_fof_pf_trade_history}')

    def get_fof_pf_amount(self, fof_id='SLW695', manager_id='py1', fund_list=WENJIAN_FUND_LIST):
        '''获取FOF私募产品金额'''
        import json
        from surfing.data.manager.manager_fof_lite import FOFDataManagerLite
        from surfing.constant import FOFTradeStatus
        try:
            def _calc_sub_asset_mv(x):
                pos = json.loads(x.position)
                df = pd.DataFrame(pos)
                if df.empty:
                    return
                df = df.set_index('fund_id')
                df = df[df.asset_type==2]
                sub_asset_mv[x.name] = df.share * df.nav
                return
            with BasicDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                    FOFInfo
                ).filter(
                    FOFInfo.fof_id.in_(fund_list),
                    FOFInfo.manager_id == manager_id,
                )
                fund_info = pd.read_sql(query.statement, query.session.bind)
                rename_dict = fund_info[["fof_id","desc_name"]].set_index("fof_id").to_dict()["desc_name"]
                rename_dict["others_mv"] = "现金"

            fof_found_date = pd.to_datetime('20200929').date()
            fof_nav = DerivedDataApi().get_fof_nav_calc(manager_id, [fof_id])
            fof_nav = fof_nav.set_index('datetime').sort_index()['nav']
            fof_sa = DerivedDataApi().get_hedge_fund_investor_pur_redemp(manager_id, [fof_id])
            fof_sa = fof_sa.set_index('datetime')
            fof_sa_sub = fof_sa.loc[fof_sa.event_type.isin([FOFTradeStatus.PURCHASE, FOFTradeStatus.SUBSCRIBE])]
            fof_sa_redemp = fof_sa.loc[fof_sa.event_type.isin([FOFTradeStatus.REDEEM,])]
            cumshares = fof_sa_sub.share_changed.fillna(0).add(fof_sa_redemp.share_changed.fillna(0), fill_value=0)
            cumshares = cumshares.groupby('datetime').sum()
            cumshares = cumshares.cumsum()
            # fof MV
            fof_mv = cumshares.reindex(fof_nav.index.union(cumshares.index)).ffill() * fof_nav
            fof_mv = fof_mv[fof_mv.index >= fof_found_date]
            pos_df = DerivedDataApi().get_fof_position(manager_id, [fof_id])
            sub_asset_mv = {}
            pos_df.set_index('datetime')[['position']].apply(_calc_sub_asset_mv, axis=1)
            # 子产品MV
            sub_asset_mv_df = pd.DataFrame.from_dict(sub_asset_mv, orient='index')
            sub_asset_mv_df = sub_asset_mv_df.reindex(fof_mv.index.union(sub_asset_mv_df.index))
            sub_asset_mv_df['others_mv'] = fof_mv - sub_asset_mv_df.sum(axis=1)
            # 子产品损益 = MV + 赎回给的现金 + 分红的现金 - 每次投入的成本
            fof_aa = FOFDataManagerLite.get_fof_asset_allocation(manager_id, [fof_id])
            fof_aa = fof_aa[fof_aa.asset_type==2].drop(columns='fof_id')
            # 赎回得到的现金
            cash_redemp = fof_aa.loc[fof_aa.event_type==FOFTradeStatus.REDEEM, ['datetime', 'fund_id', 'amount']]
            cash_redemp = cash_redemp.pivot_table(index='datetime', columns='fund_id', values='amount', aggfunc=np.sum)
            cash_redemp = cash_redemp.reindex(fof_nav.index).cumsum().ffill()
            # 成本
            cost = fof_aa.loc[fof_aa.event_type.isin([FOFTradeStatus.PURCHASE, FOFTradeStatus.SUBSCRIBE]), ['datetime', 'fund_id', 'amount']]
            cost = cost.pivot_table(index='datetime', columns='fund_id', values='amount', aggfunc=np.sum).cumsum().reindex(fof_nav.index).ffill()
            # 现金分红
            cash_dividend = fof_aa.loc[fof_aa.event_type==FOFTradeStatus.DIVIDEND_CASH, ['datetime', 'fund_id', 'amount']]
            if cash_dividend.empty:
                cash_dividend = None
            else:
                cash_dividend = cost.pivot_table(index='datetime', columns='fund_id', values='amount', aggfunc=np.sum).cumsum().reindex(fof_nav.index).ffill()
            if cash_dividend is not None:
                sub_asset_pal_fixed = sub_asset_mv_df.add(cash_redemp.fillna(0), fill_value=0).add(cash_dividend.fillna(0), fill_value=0).sub(cost.fillna(0), fill_value=0)
            else:
                sub_asset_pal_fixed = sub_asset_mv_df.add(cash_redemp.fillna(0), fill_value=0).sub(cost.fillna(0), fill_value=0)
            # FOF MV(周度)
            fof_mv_resampled = fof_mv.set_axis(pd.to_datetime(fof_mv.index), axis=0)
            fof_mv_resampled = fof_mv_resampled.resample('1W').last()
            fof_mv_resampled = fof_mv_resampled.set_axis(fof_mv_resampled.index.date, axis=0)
            # 子产品MV(周度)
            sub_asset_mv_resampled = sub_asset_mv_df.set_axis(pd.to_datetime(sub_asset_mv_df.index), axis=0)
            sub_asset_mv_resampled = sub_asset_mv_resampled.resample('1W').last()
            sub_asset_mv_resampled = sub_asset_mv_resampled.set_axis(sub_asset_mv_resampled.index.date, axis=0)
            # 子产品损益(周度)
            sub_asset_pal_fixed_sampled = sub_asset_pal_fixed.set_axis(pd.to_datetime(sub_asset_pal_fixed.index), axis=0)
            sub_asset_pal_fixed_sampled = sub_asset_pal_fixed_sampled.resample('1W').last()
            sub_asset_pal_fixed_sampled = sub_asset_pal_fixed_sampled.set_axis(sub_asset_pal_fixed_sampled.index.date, axis=0)
            sub_asset_pal_fixed_sampled = sub_asset_pal_fixed_sampled.rename(columns = rename_dict)
            fund_mv_df = sub_asset_mv_resampled
            fund_mv_df["amount"] = fund_mv_df.sum(axis = 1)
            fund_mv_df = fund_mv_df.div(fund_mv_df.amount, axis=0).drop(columns = "amount")
            return fund_mv_df, sub_asset_mv_resampled, sub_asset_pal_fixed_sampled
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from {FOFApi.get_fof_pf_amount}')

    def _filter_nav_with_min_max_date(self, x, min_max_date):
        try:
            return x[~x.datetime.between(min_max_date.at[x.fof_id.array[0], 'min_date'], min_max_date.at[x.fof_id.array[0], 'max_date'])]
        except KeyError:
            return x

    def get_fof_nav(self, manager_id: str, fof_id_list: Tuple[str]) -> Optional[pd.DataFrame]:
        fof_nav = DerivedDataApi().get_fof_nav(manager_id, fof_id_list)
        fof_nav_public = self.get_fof_nav_public(manager_id, fof_id_list)
        if fof_nav is None or fof_nav.empty:
            fof_nav = fof_nav_public
        else:
            fof_nav = fof_nav.drop(columns=['update_time', 'create_time', 'is_deleted'])
            if fof_nav_public is not None and not fof_nav_public.empty:
                fof_nav_min_max_date = fof_nav.groupby(by='fof_id', sort=False).apply(lambda x: pd.Series({'min_date': x.datetime.min(), 'max_date': x.datetime.max()}))
                fof_nav_public = fof_nav_public.groupby(by='fof_id', sort=False, group_keys=False).apply(self._filter_nav_with_min_max_date, min_max_date=fof_nav_min_max_date)
                fof_nav = pd.concat([fof_nav, fof_nav_public], ignore_index=True).sort_values(by=['manager_id', 'fof_id', 'datetime'])
        return fof_nav

    def get_fof_nav_public(self, manager_id: str, fof_id: Tuple[str] = ()) -> Optional[pd.DataFrame]:
        from.research_api import ResearchApi

        fof_nav = DerivedDataApi().get_fof_nav_public(fof_id)
        fof_nav_tl = ResearchApi().get_pf_nav(record_cds=fof_id, source='tl')
        if fof_nav_tl is not None and not fof_nav_tl.empty:
            fof_nav_tl = fof_nav_tl[['RECORD_CD', 'END_DATE', 'NAV', 'ADJ_NAV', 'ACCUM_NAV']].rename(columns={
                    'RECORD_CD': 'fof_id',
                    'END_DATE': 'datetime',
                    'NAV': 'nav',
                    'ACCUM_NAV': 'acc_net_value',
                    'ADJ_NAV': 'adjusted_nav',
                })
        if fof_nav is None or fof_nav.empty:
            fof_nav = fof_nav_tl
        else:
            fof_nav = fof_nav.drop(columns=['update_time', 'create_time', 'is_deleted'])
            fof_nav_min_max_date = fof_nav.groupby(by='fof_id', sort=False).apply(lambda x: pd.Series({'min_date': x.datetime.min(), 'max_date': x.datetime.max()}))
            fof_nav_tl = fof_nav_tl.groupby(by='fof_id', sort=False, group_keys=False).apply(self._filter_nav_with_min_max_date, min_max_date=fof_nav_min_max_date)
            fof_nav = pd.concat([fof_nav, fof_nav_tl], ignore_index=True).sort_values(by=['fof_id', 'datetime'])
        fof_nav['manager_id'] = manager_id
        return fof_nav




from typing import Dict, Any, Tuple, List
import math
import json
import io
import re
import datetime

import numpy as np
import pandas as pd

from sqlalchemy.orm import sessionmaker

from ...util.singleton import Singleton
from ...constant import FOFTradeStatus, DividendCalcMethod, EMailParserType, FOFIncentiveFeeMode
from ..view.derived_models import HedgeFundInvestorPurAndRedemp, HedgeFundInvestorDivAndCarry, HedgeFundCustodianData
from ..wrapper.mysql import DerivedDatabaseConnector
from ..api.basic import BasicDataApi
from ..api.derived import DerivedDataApi
from ..nav_reader.parser_runner import ParserRunner


class HedgeFundDataManager(metaclass=Singleton):

    def __init__(self):
        pass

    def _get_total_share(self, manager_id: str, fof_id: str, date: str, investor_id: str = '', exclude1: List[int] = [], exclude2: List[str] = []) -> Tuple[float, pd.DataFrame]:
        date = pd.to_datetime(date, infer_datetime_format=True).date()

        part1 = DerivedDataApi().get_hedge_fund_investor_pur_redemp(manager_id, [fof_id])
        assert part1 is not None, f'get pur redemp info failed (fof_id){fof_id} (manager_id){manager_id}'
        part1 = part1[part1.datetime <= date]
        part1 = part1[~part1.id.isin(exclude1)]
        part1_total_share = part1.share_changed.sum()
        if investor_id:
            part1 = part1.loc[part1.investor_id == investor_id, :]

        part2 = DerivedDataApi().get_hedge_fund_investor_div_carry(manager_id, [fof_id])
        assert part2 is not None, f'get div carry info failed (fof_id){fof_id} (manager_id){manager_id}'
        part2 = part2[part2.datetime <= date]
        part2 = part2[~part2.id.isin(exclude2)]
        part2_total_share = part2.share_changed.sum()
        if investor_id:
            part2 = part2.loc[part2.investor_id == investor_id, :]

        total_share = part1_total_share + part2_total_share
        share_separate = part1[['investor_id', 'share_changed']].groupby(by='investor_id', sort=False).sum().add(part2[['investor_id', 'share_changed']].groupby(by='investor_id', sort=False).sum(), fill_value=0)
        return total_share, share_separate

    def _refresh_datas_after_updating(self, manager_id: str, fof_id: str, investor_id: str, datetime):
        datetime: datetime.date = pd.to_datetime(datetime, infer_datetime_format=True).date()

        part1 = DerivedDataApi().get_hedge_fund_investor_pur_redemp(manager_id, [fof_id])
        assert part1 is not None, f'get pur redemp info failed (fof_id){fof_id} (manager_id){manager_id}'
        part1 = part1[part1.investor_id == investor_id].sort_values(by='datetime')

        part2 = DerivedDataApi().get_hedge_fund_investor_div_carry(manager_id, [fof_id])
        assert part2 is not None, f'get div carry info failed (fof_id){fof_id} (manager_id){manager_id}'
        part2 = part2[part2.investor_id == investor_id]

        total = pd.concat([part1, part2]).sort_values(by='datetime')
        total = total[total.datetime <= datetime]
        if total.empty:
            last_valid_share: float = 0
        else:
            last_valid_share: float = round(total.iloc[-1, :].share_after_trans, 2)

        part1_shares = part1.loc[part1.datetime > datetime, :]
        if not part1_shares.empty:
            part1_shares = part1_shares.set_index('datetime')
            part2_shares = part2.loc[part2.datetime > datetime, :]
            if not part2_shares.empty:
                part2_shares = part2_shares.set_index('datetime')
                part2_shares = part2_shares.reindex(part1_shares.index.union(part2_shares.index), method='ffill').reindex(part1_shares.index)
                part1_shares['share_changed'] = part1_shares.share_changed.add(part2_shares.share_changed, fill_value=0)

            to_refresh = part1_shares.set_index('id').share_changed.cumsum().round(2) + last_valid_share

            Session = sessionmaker(DerivedDatabaseConnector().get_engine())
            db_session = Session()
            for row in db_session.query(HedgeFundInvestorPurAndRedemp).filter(
                HedgeFundInvestorPurAndRedemp.manager_id == manager_id,
                HedgeFundInvestorPurAndRedemp.fof_id == fof_id,
                HedgeFundInvestorPurAndRedemp.investor_id == investor_id,
                HedgeFundInvestorPurAndRedemp.datetime > datetime,
            ).all():
                row.share_after_trans = float(to_refresh.at[row.id])
            db_session.commit()
            db_session.close()

    def _calc_share_details(self, fof_info, manager_id: str, fof_id: str, investor_id: str, date: str) -> List[Dict[str, Any]]:
        date = pd.to_datetime(date, infer_datetime_format=True).date()

        part1 = DerivedDataApi().get_hedge_fund_investor_pur_redemp(manager_id, [fof_id])
        assert part1 is not None, f'get pur redemp info failed (fof_id){fof_id} (manager_id){manager_id}'
        part1 = part1.loc[part1.investor_id == investor_id, ['datetime', 'event_type', 'share_changed', 'net_asset_value', 'acc_unit_value', 'water_line']]

        part2 = DerivedDataApi().get_hedge_fund_investor_div_carry(manager_id, [fof_id])
        assert part2 is not None, f'get div carry info failed (fof_id){fof_id} (manager_id){manager_id}'
        part2 = part2.loc[(part2.investor_id == investor_id) & (part2.event_type.isin([FOFTradeStatus.DIVIDEND_VOLUME, FOFTradeStatus.DEDUCT_REWARD])), ['datetime', 'event_type', 'share_changed', 'net_asset_value', 'acc_unit_value', 'water_line']]

        total = pd.concat([part1, part2]).sort_values(by='datetime')
        total = total[total.datetime <= date]
        share_details = []
        last_water_line = 1
        for row in total.itertuples(index=False):
            if row.event_type in (FOFTradeStatus.PURCHASE, FOFTradeStatus.SUBSCRIBE, FOFTradeStatus.DIVIDEND_VOLUME):
                share_details.append({'datetime': row.datetime, 'share': row.share_changed, 'water_line': row.water_line, 'acc_unit_value': row.acc_unit_value})
            elif row.event_type == FOFTradeStatus.REDEEM:
                share_changed = abs(row.share_changed)
                for one in share_details:
                    assert share_changed > 0, f'invalid share_changed {share_changed} (manager_id){manager_id} (fof_id){fof_id} (investor_id){investor_id}'
                    if (one['share'] > share_changed) or math.isclose(one['share'], share_changed, rel_tol=1e-6):
                        one['share'] -= share_changed
                        break
                    else:
                        share_changed -= one['share']
                        one['share'] = 0
                else:
                    assert False, f"not enough shares to redeem in history (manager_id){manager_id} (fof_id){fof_id} (investor_id){investor_id} (left_shares){one['share']} (left_redemp){share_changed}"
            elif row.event_type == FOFTradeStatus.DEDUCT_REWARD:
                if fof_info.incentive_fee_mode in (FOFIncentiveFeeMode.SCST_HIGH_WL_S,):
                    for one in share_details:
                        if math.isclose(one['share'], 0, rel_tol=1e-6):
                            continue
                        if one['water_line'] >= row.acc_unit_value:
                            # 当前水位线未超过该笔份额的水位线 不需要计提业绩报酬
                            continue
                        carry = round(self._calc_carry(row.acc_unit_value, one['water_line'], fof_info.incentive_fee_type, fof_info.incentive_fee_str) * one['share'] / row.net_asset_value, 2)
                        assert (one['share'] > carry) or math.isclose(one['share'], carry, rel_tol=1e-6), f"not enough shares to calc carry (manager_id){manager_id} (fof_id){fof_id} (investor_id){investor_id} (left_shares){one['share']} (carry){carry}"
                        one['share'] -= carry
                        # 水位线变成本次计提时的水位线了
                        one['water_line'] = row.acc_unit_value
                elif fof_info.incentive_fee_mode in (FOFIncentiveFeeMode.INTE_HIGH_WL, FOFIncentiveFeeMode.INTE_HIGH_WL_WITH_SCST_R):
                    assert row.water_line > last_water_line, f'water line should be increased monotonously for every investor (manager_id){manager_id} (fof_id){fof_id} (investor_id){investor_id} (last water line){last_water_line} (water line this time){row.water_line}'
                    # profit = self._calc_carry(row.acc_unit_value, last_water_line, fof_info.incentive_fee_type, fof_info.incentive_fee_str)
                else:
                    assert False, f'do not support the fee mode {fof_info.incentive_fee_mode} temp (fof_id){fof_id} (manager_id){manager_id}'
                last_water_line = row.water_line
            else:
                assert False, f'invalid event type {row.event_type} during doing carry calc'
        return share_details

    def investor_pur_redemp_update(self, datas: Dict[str, Any]) -> int:
        '''
        申赎记录更新/添加

        Parameters
        ----------
        datas : Dict[str, Any]
            数据, keys:
                id : 如果是添加 则不需要该字段; 否则则需要
                fof_id : 产品ID
                manager_id : 管理端ID
                datetime : 交易日期
                event_type : 交易类型(认购/申购/赎回)
                investor_id : 投资人ID
                net_asset_value : 单位净值
                acc_unit_value : 累计净值
                water_line : 水位线(一般是累计净值; 但整体高水位法下与累计净值可能不同)
                amount : 申购金额/认购金额
                raising_interest : 募集期利息
                redemp_share : 赎回份额
                redemp_fee : 赎回费

        Returns/Exceptions
        -------
        throw excepion if failed, or return id (an integer) if succeed
        '''
        fof_info = BasicDataApi().get_fof_info(datas['manager_id'], fof_id_list=[datas['fof_id']])
        assert fof_info is not None, '!!!!'
        assert fof_info.shape[0] == 1, f"invalid fof id {[datas['fof_id']]}"
        fof_info = fof_info.iloc[-1, :]

        event_type = datas['event_type']
        if event_type == FOFTradeStatus.PURCHASE:
            # 认购
            datas['purchase_amount'] = datas['amount']
            datas['share_changed'] = round((datas['amount'] + datas['raising_interest']) / datas['net_asset_value'], 2)
        elif event_type == FOFTradeStatus.SUBSCRIBE:
            # 申购
            datas['purchase_amount'] = datas['amount']
            datas['share_changed'] = round(datas['amount'] / datas['net_asset_value'], 2)
        elif event_type == FOFTradeStatus.REDEEM:
            # 赎回
            datas['share_changed'] = -datas['redemp_share']
            datas['redemp_confirmed_amount'] = round(datas['redemp_share'] * datas['net_asset_value'], 2) - datas['redemp_fee']

            # 计算业绩报酬计提
            carry_sum = 0
            share_details = self._calc_share_details(fof_info, datas['manager_id'], datas['fof_id'], datas['investor_id'], datas['datetime'])
            for one in share_details:
                if math.isclose(one['share'], 0, rel_tol=1e-6):
                    continue
                if (one['share'] > datas['redemp_share']) or math.isclose(one['share'], datas['redemp_share'], rel_tol=1e-6):
                    one['share'] -= datas['redemp_share']
                    if fof_info.incentive_fee_mode in (FOFIncentiveFeeMode.INTE_HIGH_WL_WITH_SCST_R,):
                        if one['acc_unit_value'] < one['water_line']:
                            # 如果这笔交易当时的交易净值在水下 需要在赎回时补扣业绩报酬
                            profit = self._calc_carry(min(datas['acc_unit_value'], one['water_line']), one['acc_unit_value'], fof_info.incentive_fee_type, fof_info.incentive_fee_str)
                            print(f"(profit){profit} (shares){datas['redemp_share']}")
                            carry_sum += round(profit * datas['redemp_share'], 2)
                    elif fof_info.incentive_fee_mode in (FOFIncentiveFeeMode.SCST_HIGH_WL_S,):
                        if datas['acc_unit_value'] > one['water_line']:
                            carry_sum += round(self._calc_carry(datas['acc_unit_value'], one['water_line'], fof_info.incentive_fee_type, fof_info.incentive_fee_str) * datas['redemp_share'], 2)
                    break
                else:
                    datas['redemp_share'] -= one['share']
                    if fof_info.incentive_fee_mode in (FOFIncentiveFeeMode.INTE_HIGH_WL_WITH_SCST_R,):
                        if one['acc_unit_value'] < one['water_line']:
                            # 如果这笔交易当时的交易净值在水下 需要在赎回时补扣业绩报酬
                            profit = self._calc_carry(min(datas['acc_unit_value'], one['water_line']), one['acc_unit_value'], fof_info.incentive_fee_type, fof_info.incentive_fee_str)
                            print(f"(profit){profit} (shares){one['share']}")
                            carry_sum += round(profit * one['share'], 2)
                    elif fof_info.incentive_fee_mode in (FOFIncentiveFeeMode.SCST_HIGH_WL_S,):
                        if datas['acc_unit_value'] > one['water_line']:
                            carry_sum += round(self._calc_carry(datas['acc_unit_value'], one['water_line'], fof_info.incentive_fee_type, fof_info.incentive_fee_str) * one['share'], 2)
                    one['share'] = 0
            else:
                assert False, f"not enough shares to redeem (fof_id){datas['fof_id']} (manager_id){datas['manager_id']} (left_shares){one['share']} (left_redemp){datas['redemp_share']}"
            datas['carry_amount'] = carry_sum
            datas['redemp_confirmed_amount'] = round(datas['redemp_confirmed_amount'] - carry_sum, 2)
        else:
            raise ValueError(f'invalid event_type {event_type}')

        _, share_separate = self._get_total_share(datas['manager_id'], datas['fof_id'], datas['datetime'], datas['investor_id'], exclude1=[datas['id']] if 'id' in datas else [])

        datas['share_changed'] = round(datas['share_changed'], 2)
        try:
            datas['share_after_trans'] = round(float(share_separate.at[datas['investor_id'], 'share_changed']) + datas['share_changed'], 2)
        except KeyError:
            datas['share_after_trans'] = datas['share_changed']

        datas['applied_date'] = datas['datetime']
        datas['deposited_date'] = datas['datetime']
        if 'water_line' not in datas:
            datas['water_line'] = datas['acc_unit_value']

        s = pd.Series(datas).drop(['amount', 'redemp_share'], errors='ignore')
        # TODO: 原子操作下面几步
        # TODO: 先刷后边的再删这一条
        if 'id' in datas:
            DerivedDataApi().delete_hedge_fund_investor_pur_redemp(id_to_delete=datas['id'])
        sql_obj = HedgeFundInvestorPurAndRedemp()
        for k, v in s.to_dict().items():
            setattr(sql_obj, k, v)
        with DerivedDatabaseConnector().managed_session() as session:
            session.add(sql_obj)
            session.commit()
            the_id = sql_obj.id
        # df.to_sql(HedgeFundInvestorPurAndRedemp.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
        self._refresh_datas_after_updating(datas['manager_id'], datas['fof_id'], datas['investor_id'], datas['datetime'])
        return the_id

    def _calc_carry(self, nav: float, water_line: float, fee_type: str, fee_desc: str) -> float:
        profit: float = nav - water_line
        if fee_type == '1':
            return profit * float(fee_desc)
        elif fee_type == '2':
            fee_df = pd.DataFrame(json.loads(fee_desc)).sort_values(by='start', ascending=False)
            assert not fee_df.empty, f'invalid fee str {fee_desc} (fee_type){fee_type}'
            total_profit = 0
            for row in fee_df.itertuples(index=False):
                assert profit > 0, f'invalid profit (profit){profit} (nav){nav} (water_line){water_line} (fee_desc){fee_desc} (fee_type){fee_type}'
                level_profit: float = water_line * (row.start if row.start else 0)
                if profit > level_profit:
                    total_profit += (profit - level_profit) * (row.val if row.val else 0)
                    profit = level_profit
            return total_profit
        else:
            assert False, f'invalid fee type {fee_type}, do not support to calc carry by it'

    def investor_pur_redemp_remove(self, id: int):
        '''
        申赎记录删除

        Parameters
        ----------
        id : int
            对应数据库中id

        Returns/Exceptions
        -------
        throw excepion if failed
        '''

        df = DerivedDataApi().get_hedge_fund_investor_pur_redemp_by_id(id)
        assert df is not None, f'failed to get hedge fund investor data by id {id}'
        assert df.shape[0] == 1, f'invalid id {id}'

        s = df.iloc[-1, :]
        DerivedDataApi().delete_hedge_fund_investor_pur_redemp(id_to_delete=id)
        self._refresh_datas_after_updating(s.manager_id, s.fof_id, s.investor_id, s.datetime)

    def investor_div_carry_add(self, datas: Dict[str, Any]):
        '''
        分红/业绩计提记录添加

        Parameters
        ----------
        datas : Dict[str, Any]
            数据, keys:
                fof_id : 产品ID
                datetime : 分红/计提日期
                manager_id : 管理端ID
                event_type : 交易类型(现金分红/红利再投/计提业绩报酬)
                investor_id : 投资人ID
                net_asset_value : 单位净值
                acc_unit_value : 累计净值
                water_line : 水位线(一般是累计净值; 但整体高水位法下与累计净值可能不同)
                total_dividend : 红利总额
                cash_dividend : 现金分红金额
                reinvest_amount : 再投资金额
                carry_amount : 业绩报酬金额
                share_changed : 份额变更
                share_after_trans : 分红/计提后份额

        Returns/Exceptions
        -------
        throw excepion if failed, or return id (an integer) if succeed
        '''

        event_type = datas['event_type']
        if event_type == FOFTradeStatus.DEDUCT_REWARD:
            # 计提业绩报酬 暂不需要额外处理
            datas['share_changed'] = round(datas['share_changed'], 2)
            datas['share_after_trans'] = round(datas['share_after_trans'], 2)
        else:
            total_share, share_separate = self._get_total_share(datas['manager_id'], datas['fof_id'], datas['datetime'], datas['investor_id'])
            share_separate = share_separate.share_changed.array[-1]
            # div = datas['total_dividend'] * share_separate / total_share

            div = datas['total_dividend']
            if event_type == FOFTradeStatus.DIVIDEND_CASH:
                # 现金分红
                datas['share_changed'] = 0
                datas['cash_dividend'] = div - (datas['carry_amount'] if not pd.isnull(datas['carry_amount']) else 0)
                datas['reinvest_amount'] = 0
            elif event_type == FOFTradeStatus.DIVIDEND_VOLUME:
                # 红利再投
                datas['reinvest_amount'] = div - (datas['carry_amount'] if not pd.isnull(datas['carry_amount']) else 0)
                datas['share_changed'] = round(datas['reinvest_amount'] / datas['net_asset_value'], 2)
                datas['cash_dividend'] = 0
            else:
                raise ValueError(f'invalid event_type {event_type}')

            datas['share_after_trans'] = round(share_separate + datas['share_changed'], 2)

            for k in ('reinvest_amount', 'cash_dividend', 'total_dividend'):
                datas[k] = float(datas[k])

        for k in ('net_asset_value', 'acc_unit_value', 'water_line', 'carry_amount', 'share_changed', 'share_after_trans'):
            datas[k] = float(datas[k])

        datas['applied_date'] = datas['datetime']
        datas['deposited_date'] = datas['datetime']
        df = pd.Series(datas).to_frame().T
        # TODO: 原子操作下面几步
        if 'id' in datas:
            DerivedDataApi().delete_hedge_fund_investor_div_carry(id_to_delete=datas['id'])
        df.to_sql(HedgeFundInvestorDivAndCarry.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')

    def investor_div_carry_remove(self, id: int):
        '''
        分红/业绩计提记录删除

        Parameters
        ----------
        id : int
            对应数据库中id

        Returns/Exceptions
        -------
        throw excepion if failed
        '''

        df = DerivedDataApi().get_hedge_fund_investor_div_carry_by_id(id)
        assert df is not None, f'failed to get hedge fund investor data by id {id}'
        assert df.shape[0] == 1, f'invalid id {id}'

        s = df.iloc[-1, :]
        DerivedDataApi().delete_hedge_fund_investor_div_carry(id_to_delete=id)
        self._refresh_datas_after_updating(s.manager_id, s.fof_id, s.investor_id, s.datetime)

    def calc_dividend_event(self, datas: Dict[str, Any]) -> pd.DataFrame:
        '''
        计算分红事件

        Parameters
        ----------
        datas : Dict[str, Any]
            数据, keys:
                fof_id : 产品ID
                manager_id : 管理端ID
                datetime : 分红日期
                event_type : 红利发放方式(现金分红/红利再投)
                net_asset_value : 单位净值
                acc_unit_value : 累计净值
                calc_method : 计算方法(红利总额/每单位红利)
                dividend_for_calc : 红利总额/每单位红利
                do_deduct_reward: 是否计提业绩报酬

        Returns/Exceptions
        -------
        throw excepion if failed, or return a pd.DataFrame if succeed
        '''

        total_share, share_separate = self._get_total_share(datas['manager_id'], datas['fof_id'], datas['datetime'])
        share_separate = share_separate.share_changed
        weight = share_separate / total_share

        calc_method = datas['calc_method']
        if calc_method == DividendCalcMethod.BY_TOTAL_AMOUNT:
            # 红利总额
            dividend_amount = datas['dividend_for_calc']
        elif calc_method == DividendCalcMethod.BY_PER_UNIT:
            # 每单位红利
            dividend_amount = round(total_share * datas['dividend_for_calc'], 2)
        else:
            raise ValueError(f'invalid calc_method {calc_method}')

        s = (dividend_amount * weight).round(2)
        event_type = datas['event_type']
        if datas['event_type'] == FOFTradeStatus.DIVIDEND_CASH:
            # 现金分红
            df = s.to_frame('total_dividend')
            df['cash_dividend'] = df.total_dividend
            if 'do_deduct_reward' in datas and datas['do_deduct_reward']:
                carry_df = self.calc_carry_event(datas)
                df = df.join(carry_df.set_index('investor_id').carry_amount)
                df['cash_dividend'] -= df.carry_amount.round(2)
            else:
                df['carry_amount'] = np.nan
            df['share_changed'] = 0
            df['reinvest_amount'] = 0
        elif datas['event_type'] == FOFTradeStatus.DIVIDEND_VOLUME:
            # 红利再投
            df = s.to_frame('total_dividend')
            df['reinvest_amount'] = df.total_dividend
            if 'do_deduct_reward' in datas and datas['do_deduct_reward']:
                carry_df = self.calc_carry_event(datas)
                df = df.join(carry_df.set_index('investor_id').carry_amount)
                df['reinvest_amount'] -= df.carry_amount.round(2)
            else:
                df['carry_amount'] = np.nan
            df['share_changed'] = (df.reinvest_amount / datas['net_asset_value']).round(2)
            df['cash_dividend'] = 0
        else:
            raise ValueError(f'invalid event_type {event_type}')
        df['share_after_trans'] = (share_separate + df.share_changed).round(2)

        df['fof_id'] = datas['fof_id']
        df['manager_id'] = datas['manager_id']
        df['datetime'] = datas['datetime']
        df['event_type'] = datas['event_type']
        columns_in_df = set(df.columns.array)
        if 'net_asset_value' not in columns_in_df:
            df['net_asset_value'] = datas['net_asset_value']
        if 'acc_unit_value' not in columns_in_df:
            df['acc_unit_value'] = datas['acc_unit_value']
        if 'water_line' not in columns_in_df:
            df['water_line'] = datas['acc_unit_value']
        return df.reset_index()

    def calc_carry_event(self, datas: Dict[str, Any]) -> pd.DataFrame:
        '''
        计算业绩计提事件

        Parameters
        ----------
        datas : Dict[str, Any]
            数据, keys:
                fof_id : 产品ID
                manager_id : 管理端ID
                datetime : 计提日期
                net_asset_value : 单位净值
                acc_unit_value : 累计净值

        Returns/Exceptions
        -------
        throw excepion if failed, or return a pd.DataFrame if succeed
        '''

        fof_info = BasicDataApi().get_fof_info(datas['manager_id'], fof_id_list=[datas['fof_id']])
        assert fof_info is not None, f"get fof info failed (fof_id){datas['fof_id']} (manager_id){datas['manager_id']}"
        assert fof_info.shape[0] == 1, f"invalid fof id {[datas['fof_id']]}"
        fof_info = fof_info.iloc[-1, :]

        date = pd.to_datetime(datas['datetime'], infer_datetime_format=True).date()

        part1 = DerivedDataApi().get_hedge_fund_investor_pur_redemp(datas['manager_id'], [datas['fof_id']])
        assert part1 is not None, f"get pur redemp info failed (fof_id){datas['fof_id']} (manager_id){datas['manager_id']}"
        part1 = part1[part1.datetime <= date]

        part2 = DerivedDataApi().get_hedge_fund_investor_div_carry(datas['manager_id'], [datas['fof_id']])
        assert part2 is not None, f"get pur div carry info failed (fof_id){datas['fof_id']} (manager_id){datas['manager_id']}"
        part2 = part2[part2.datetime <= date]

        carry_data = []
        total = pd.concat([part1, part2])
        if fof_info.incentive_fee_mode in (FOFIncentiveFeeMode.INTE_HIGH_WL, FOFIncentiveFeeMode.INTE_HIGH_WL_WITH_SCST_R):
            deduct_reward_info = part2[part2.event_type == FOFTradeStatus.DEDUCT_REWARD]
            if deduct_reward_info.empty:
                last_water_line = 1
            else:
                last_water_line = deduct_reward_info.sort_values(by='datetime').acc_unit_value.array[-1]
            print(f"(fof_id){datas['fof_id']} (manager_id){datas['manager_id']} (last_water_line){last_water_line} (acc_unit_value){datas['acc_unit_value']}")
            # 如果产品的水位线变高了 计提业绩报酬
            assert datas['acc_unit_value'] > last_water_line, f"latest acc nav {last_water_line} is not greater than the last water line {datas['acc_unit_value']}, can not do carry calc (fof_id){datas['fof_id']} (manager_id){datas['manager_id']}"
            profit = self._calc_carry(datas['acc_unit_value'], last_water_line, fof_info.incentive_fee_type, fof_info.incentive_fee_str)
            for investor_id in total.investor_id.unique():
                share_details = self._calc_share_details(fof_info, datas['manager_id'], datas['fof_id'], investor_id, datas['datetime'])
                share_before_trans = round(pd.DataFrame(share_details).share.sum(), 2)
                carry_amount_sum = 0
                for one in share_details:
                    if math.isclose(one['share'], 0, rel_tol=1e-6):
                        continue
                    if one['datetime'] == date:
                        # 今天新买入的交易不做业绩计提
                        continue
                    carry_amount_sum += round(profit * one['share'], 2)
                if not math.isclose(carry_amount_sum, 0):
                    carry_data.append({
                        'investor_id': investor_id,
                        'datetime': datas['datetime'],
                        'share_before_trans': share_before_trans,
                        'carry_amount': carry_amount_sum,
                        'share_changed': 0,
                        'share_after_trans': share_before_trans,
                        'net_asset_value': round(datas['net_asset_value'] - profit, fof_info.nav_decimals),
                        'acc_unit_value': round(datas['acc_unit_value'] - profit, fof_info.nav_decimals),
                        'water_line': round(datas['acc_unit_value'] - profit, fof_info.nav_decimals),
                    })
        elif fof_info.incentive_fee_mode in (FOFIncentiveFeeMode.SCST_HIGH_WL_S,):
            for investor_id in total.investor_id.unique():
                share_details = self._calc_share_details(fof_info, datas['manager_id'], datas['fof_id'], investor_id, datas['datetime'])
                share_before_trans = round(pd.DataFrame(share_details).share.sum(), 2)
                carry_amount_sum = 0
                share_changed = 0
                for one in share_details:
                    if math.isclose(one['share'], 0, rel_tol=1e-6):
                        continue
                    if one['water_line'] >= datas['acc_unit_value']:
                        # 当前水位线未超过该笔份额的水位线 不需要计提业绩报酬
                        continue
                    carry_amount = round(self._calc_carry(datas['acc_unit_value'], one['water_line'], fof_info.incentive_fee_type, fof_info.incentive_fee_str) * one['share'], 2)
                    carry = round(carry_amount / datas['net_asset_value'], 2)
                    assert (one['share'] > carry) or math.isclose(one['share'], carry, rel_tol=1e-6), f"not enough shares to calc carry (fof_id){datas['fof_id']} (manager_id){datas['manager_id']} (left_share){one['share']} (carry){carry}"
                    carry_amount_sum += carry_amount
                    share_changed -= carry
                    one['share'] -= carry
                    # 水位线变成本次计提时的水位线了
                    one['water_line'] = datas['acc_unit_value']
                carry_data.append({
                    'investor_id': investor_id,
                    'datetime': datas['datetime'],
                    'share_before_trans': share_before_trans,
                    'carry_amount': carry_amount_sum,
                    'share_changed': share_changed,
                    'share_after_trans': share_before_trans + share_changed,
                    'net_asset_value': datas['net_asset_value'],
                    'acc_unit_value': datas['acc_unit_value'],
                    'water_line': datas['acc_unit_value'],
                })
        else:
            assert False, f"do not support the fee mode {fof_info.incentive_fee_mode} temp (fof_id){datas['fof_id']} (manager_id){datas['manager_id']}"

        df = pd.DataFrame(carry_data)
        df['fof_id'] = datas['fof_id']
        df['manager_id'] = datas['manager_id']
        df['datetime'] = datas['datetime']
        df['event_type'] = FOFTradeStatus.DEDUCT_REWARD
        return df

    def upload_custodian_data(self, manager_id: str, fof_id: str, datas: io.BytesIO) -> datetime.date:
        '''
        上传/更新私募基金托管数据

        Parameters
        ----------
        manager_id : str
            manager id
        fof_id : str
            fof id
        datas : io.BytesIO
            open 托管数据表格获得的 file object

        Returns/Exceptions
        -------
        throw excepion if failed, or return a datetime.date if succeed
        '''

        parser_runner = ParserRunner(manager_id, EMailParserType.E_PARSER_VALUATION)
        total_df = parser_runner.parse_data(datas, manager_id)
        if total_df is None:
            raise Exception(f'could not parse this custodian data (manager_id){manager_id} (fof_id){fof_id}')

        total_df['fof_id'] = fof_id
        now_df = DerivedDataApi().get_hedge_fund_custodian_data(manager_id=manager_id, fof_id_list=[fof_id])
        if now_df is not None:
            now_df = now_df.drop(columns=['update_time', 'create_time', 'is_deleted', 'id', 'url']).replace({None: np.nan}).astype(total_df.dtypes.to_dict())
            # merge on all columns
            df = total_df.round(6).merge(now_df.round(6), how='left', indicator=True, validate='one_to_one')
            df = df[df._merge == 'left_only'].drop(columns=['_merge'])
        else:
            df = total_df
        if not df.empty:
            print(df)
            for row in df.itertuples(index=False):
                DerivedDataApi().delete_hedge_fund_custodian_data(row.manager_id, row.fof_id, row.datetime)
            df.to_sql(HedgeFundCustodianData.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
        return total_df['datetime'].array[0]

    @staticmethod
    def calc_whole_adjusted_net_value(nav_acc_data: pd.DataFrame) -> pd.DataFrame:
        '''
        计算单只基金的全量复权净值

        Parameters
        ----------
        nav_acc_data : pd.DataFrame
            净值数据, index: datetime.date 净值日期
                columns:
                    net_asset_value : 单位净值(必须从1开始, 除非与累计净值完全相同)
                    acc_unit_value : 累计净值(必须从1开始, 除非与单位净值完全相同)

        Returns/Exceptions
        -------
        throw excepion if failed, or return a pd.DataFrame if succeed

        value returned : pd.DataFrame
            复权净值数据, index: datetime.date 净值日期
                columns:
                    adj_nav : 复权净值
                    adj_factor: 复权因子
                    change_rate: 涨跌幅
        '''

        # 先按index(datetime.date)排序
        nav_acc_data: pd.DataFrame = nav_acc_data.sort_index()
        if nav_acc_data.net_asset_value.equals(nav_acc_data.acc_unit_value):
            # 如果单位净值和累计净值完全一样 允许不从1开始 走一个单独的(简易)计算流程
            df = nav_acc_data
            df['adj_nav'] = df.net_asset_value
            df['change_rate'] = df.adj_nav.pct_change()
            df['adj_factor'] = 1
            df = df[['adj_nav', 'adj_factor', 'change_rate']]
            return df

        assert math.isclose(nav_acc_data.net_asset_value.array[0], 1) and math.isclose(nav_acc_data.acc_unit_value.array[0], 1), f'fund nav data should start with 1 (data){nav_acc_data}'

        assert nav_acc_data.acc_unit_value.round(6).ge(nav_acc_data.net_asset_value.round(6)).all(), f'acc_unit_value should always ge than net_asset_value (data){nav_acc_data}'
        # 公式一
        # days_before_dividend: pd.Series = (nav_acc_data.acc_unit_value - nav_acc_data.net_asset_value).round(6).drop_duplicates(keep='last')
        # 公式二
        days_before_dividend: pd.Series = (nav_acc_data.acc_unit_value - nav_acc_data.net_asset_value).round(6).drop_duplicates(keep='first')
        assert days_before_dividend.notna().all(), f'fund nav data should not contains NaN (data){nav_acc_data}'
        # 这里算出来每天的分红数据不应该出现负值 这样前边的 drop_duplicates 才是对的
        assert days_before_dividend.is_monotonic_increasing, f'dividend should always be monotonic increasing (data){nav_acc_data}'

        divdidend_that_day: pd.Series = days_before_dividend.sub(days_before_dividend.shift(1, fill_value=0))
        # 公式一
        # adj_factor: pd.Series = 1 + divdidend_that_day / (nav_acc_data.loc[divdidend_that_day.index, 'net_asset_value'].shift(1, fill_value=1) - divdidend_that_day)
        # 公式二
        net_asset_value = nav_acc_data.loc[divdidend_that_day.index, 'net_asset_value']
        adj_factor: pd.Series = (net_asset_value + divdidend_that_day) / net_asset_value

        # 公式一
        # df: pd.DataFrame = adj_factor.cumprod().reindex(nav_acc_data.index, method='bfill').to_frame('adj_factor')
        # 公式二
        df: pd.DataFrame = adj_factor.cumprod().reindex(nav_acc_data.index, method='ffill').to_frame('adj_factor')
        df['adj_nav'] = df.adj_factor * nav_acc_data.net_asset_value
        df['change_rate'] = df.adj_nav.pct_change()
        df = df[['adj_nav', 'adj_factor', 'change_rate']]
        return df
